/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Copyright 2011 Rob Smits, Divam Jain, Sarah Pidcock
 */

// Standard C++ libraries
#include <iostream>
#include <iomanip>
#include <sstream>
#include <map>

// C TCP/IP libraries
#include <netinet/in.h>
#include <netinet/tcp.h>   //Provides declarations for tcp header
#include <netinet/ip.h>    //Provides declarations for ip header
#include <netdb.h>
#include <arpa/inet.h>

// Misc C libraries
#include <time.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <pthread.h>

#include <linux/netfilter.h>  /* Defines verdicts (NF_ACCEPT, etc) */
#include <libnetfilter_queue/libnetfilter_queue.h>

//BridgeSPA includes
#include "crypto.c"
#include "utils.cpp"
#include "tcpsession.h"

using namespace std;

#define CONTROL_PORT 34553

//Map of <IP, port> to sequence number offset
map<TcpSession, pair<int32_t, uint32_t> > spa_data;

map<TcpSession, unsigned char> old_timestamps;

//Map of <IP, port> to key data
map<pair<uint32_t, uint16_t>, string> bridge_keys;

pthread_mutex_t bridge_entries_mutex = PTHREAD_MUTEX_INITIALIZER;
 
int parseBridgeEntry(char buf[1024], uint32_t &ip, uint16_t &port, string &bridge_key) {
  string sbuf;
  stringstream ss (stringstream::in | stringstream::out);
  struct sockaddr_in bridge;
  
  ss << buf; 
  sbuf = ss.str(); 
  
  uint16_t colon_pos = sbuf.find(":");
  uint16_t bar_pos = sbuf.find("|");
  
  if (colon_pos <= 0 || bar_pos <= 0) return 0;

  string s_ip = sbuf.substr(0, colon_pos);
  string s_port = sbuf.substr(colon_pos + 1, bar_pos);
  bridge_key = sbuf.substr(bar_pos+1);

  inet_aton(s_ip.c_str(), &bridge.sin_addr);

  ip = (uint32_t)ntohl(bridge.sin_addr.s_addr);
  
  ss.clear();//clear any bits set
  ss.str(string());
  ss << s_port;
  ss >> port;
  
  return 1;
}

void addBridgeEntry(uint32_t ip, uint16_t port, string bridge_key) {
  pthread_mutex_lock( &bridge_entries_mutex);
  
  pair<uint32_t, uint16_t> ipport;
  ipport.first = ip;
  ipport.second = port;
  
  bridge_keys[ipport] = bridge_key;
  
  pthread_mutex_unlock( &bridge_entries_mutex);
}

string getBridgeEntry(uint32_t ip, uint16_t port) {
  pthread_mutex_lock( &bridge_entries_mutex);
  
  pair<uint32_t, uint16_t> ipport;
  ipport.first = ip;
  ipport.second = port;
  
  string result = bridge_keys[ipport];
  
  pthread_mutex_unlock( &bridge_entries_mutex);
  return result;
}

void removeQueueRule(uint32_t ip, uint16_t port) {
  stringstream ss (stringstream::in | stringstream::out);
  string cmd;
  int result = 0;


  struct sockaddr_in bridge;
  bridge.sin_addr.s_addr = htonl(ip);

  char* a_ip = inet_ntoa(bridge.sin_addr);

  ss << "iptables -D OUTPUT -p tcp -d " << a_ip << " --dport " << port << " -j NFQUEUE --queue-num 1";
  cout << "About to execute: " << ss.str() << endl;
  result = system(ss.str().c_str());

  ss.clear();
  ss.str(string());

  if (result > 0) {
    cout << "There was a problem executing the command\n" << endl;
  }

  ss << "iptables -D INPUT -p tcp -s " << a_ip << " --sport " << port << " -j NFQUEUE --queue-num 1";
  cout << "About to execute: " << ss.str() << endl;
  result = system(ss.str().c_str());

  if (result > 0) {
    cout << "There was a problem executing the command\n" << endl;
  }
}

void removeQueueRulesAndDie(int sig) {
  pthread_mutex_lock( &bridge_entries_mutex);

  map<pair<uint32_t, uint16_t>, string>::iterator it;

  for (it=bridge_keys.begin(); it != bridge_keys.end(); it++ ) {
    pair<uint32_t, uint16_t> ipport = it->first;
    removeQueueRule(ipport.first, ipport.second);
  }

  pthread_mutex_unlock( &bridge_entries_mutex);
  exit(1);
}

void addQueueRule(uint32_t ip, uint16_t port) {
  stringstream ss (stringstream::in | stringstream::out);
  string cmd;
  int result = 0;

  struct sockaddr_in bridge;
  bridge.sin_addr.s_addr = htonl(ip);

  char* a_ip = inet_ntoa(bridge.sin_addr);

  ss << "iptables -I OUTPUT -p tcp -d " << a_ip << " --dport " << port << " -j NFQUEUE --queue-num 1";
  cout << "About to execute: " << ss.str() << endl;
  result = system(ss.str().c_str());

  ss.clear();
  ss.str(string());

  if (result > 0) {
    cout << "There was a problem executing the command\n" << endl;
  }

  ss << "iptables -I INPUT -p tcp -s " << a_ip << " --sport " << port << " -j NFQUEUE --queue-num 1";
  cout << "About to execute: " << ss.str() << endl;
  result = system(ss.str().c_str());

  if (result > 0) {
    cout << "There was a problem executing the command\n" << endl;
  }
}

void handleControlPortClient(int clientfd) {
  uint32_t bytes_recvd;
  char buf[1024];
  uint32_t ip;
  uint16_t port;
  string bridge_key;
  
  if ((bytes_recvd = recv(clientfd, buf, sizeof buf, 0)) <= 0) {
    cout << "Error reading from client: " << clientfd << endl;
    close(clientfd);
	return;
  }
  
  close(clientfd);
  
  if (!parseBridgeEntry(buf, ip, port, bridge_key)) {
    cout << "Error parsing data from client: " << buf << endl;
    return;
  }
  
  cout << "Read ip port key " << ip << " " << port << " " << bridge_key << endl;
  
  addQueueRule(ip, port);
  addBridgeEntry(ip, port, bridge_key);
}

void* serveControlPort(void* data) {
  int serverfd;
  struct sockaddr_in server_in;

  if ((serverfd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
    Die("Failed to create socket");
  }

  memset(&server_in, 0, sizeof(server_in));
  server_in.sin_family = AF_INET;
  server_in.sin_addr.s_addr = htonl(INADDR_ANY);
  server_in.sin_port = htons(CONTROL_PORT);

  if (bind(serverfd, (struct sockaddr *) &server_in, sizeof(server_in)) < 0) {
    Die("Failed to bind the server socket");
  }
  

  if (listen(serverfd, 5) < 0) {
    Die("Failed to listen on server socket");
  }

  while (1) {
    int clientfd = 0;
	unsigned int clientlen = sizeof(struct sockaddr_in);
	struct sockaddr_in client_in;
	
	memset((void*)&client_in, 0, sizeof(struct sockaddr_in));
	
	
    if ((clientfd = accept(serverfd, (struct sockaddr *) &client_in, &clientlen)) < 0) {
      Die("Failed to accept client connection");
    }
    printf("Client connected: %s\n", inet_ntoa(client_in.sin_addr));
    handleControlPortClient(clientfd);
  }
  
  return 0;
}

pair<int32_t, uint32_t> findSpaData(TcpSession to_find) {
  //Accessing the map with the array accessor will create an empty entry if
  //an entry is not found. This is not an issue if we assume that will always
  //be followed by populating and using that empty element.
  //Note that this may cause a bug if the offset really is zero 
  return spa_data[to_find];
}

void setSpaData(TcpSession to_set, int32_t offset, uint32_t conn_tag) {
  pair<int32_t, uint32_t> new_value;
  new_value.first = offset;
  new_value.second = conn_tag;

  spa_data[to_set] = new_value;
}


void sleep_timestamp(unsigned char old_byte, unsigned char new_byte) {
  if (new_byte >= old_byte) {
    msleep(10*(new_byte - old_byte));
  } else {
    msleep(10*(0x0100 + new_byte - old_byte));
  }
}


void setTimestamp(struct tcphdr *tcp_hdr, unsigned char new_byte, bool ts_echo, TcpSession session) {
  unsigned char *options_ptr = (unsigned char *)(tcp_hdr) + 20;
  uint16_t data_offset = *((uint8_t*)((unsigned char *)(tcp_hdr) + 12));
  uint32_t *long_ptr;
  uint32_t old_timestamp = 0;
  uint32_t new_timestamp = 0;

  data_offset = data_offset >> 4;  //only first 4 bits
  data_offset *= 4;  

  cout << hex;

  while(1) {
    uint8_t kind = (uint8_t)(*options_ptr);
    options_ptr++;
    
    if (kind == 1) {
      continue;
    }

    uint8_t len = (uint8_t)*options_ptr;
    options_ptr++;

    if (kind == 8) {
      unsigned char old_byte = 0;

      long_ptr = (uint32_t*) options_ptr;
      old_timestamp = ntohl(*long_ptr);
      cout << "Timestamp value: " << old_timestamp << endl;
      
     
      if (!ts_echo) {
        old_byte = ((unsigned char *)(&old_timestamp))[0];
        new_timestamp = old_timestamp;
        ((unsigned char *)(&new_timestamp))[0] = new_byte;
        if (new_byte < old_byte) new_timestamp += 0x0100;
        *long_ptr = htonl(new_timestamp);
        sleep_timestamp(old_byte, new_byte);
        cout << "NEW Timestamp value: " << new_timestamp << endl;
        old_timestamps[session] = old_byte;
      } 

      options_ptr += 4;
      long_ptr = (uint32_t*) options_ptr;

      if (ts_echo) {
        unsigned char old_byte_last = old_timestamps[session]; //should check that this item existed first
        old_timestamp = ntohl(*long_ptr);
        cout << "Echo value: " << old_timestamp << endl;
        new_timestamp = old_timestamp;
        cout << "New echo byte value: " << new_byte << endl;
        ((unsigned char *)(&new_timestamp))[0] = old_byte_last;
        if (new_byte < old_byte_last) new_timestamp -= 0x0100;
        *long_ptr = htonl(new_timestamp);
        cout << "NEW echo value: " << new_timestamp << endl;
      }

      options_ptr += 4;

    } else if (kind == 0) {
      break;
    } else {
      options_ptr += len - 2;
    }

    if (((uint64_t)options_ptr) - ((uint64_t)tcp_hdr) >= data_offset) {
      break;
    }
  }

  cout << dec;
}

void getTagData(TcpSession session, int32_t &spa_offset, uint32_t &spa_conn_tag, struct iphdr *ip_hdr, struct tcphdr *tcp_hdr) {
  string key;
  SharedSecret bridgeSecret;

  struct sockaddr_in source_ip, dest_ip;
  source_ip.sin_addr.s_addr = ip_hdr->saddr;
  dest_ip.sin_addr.s_addr = ip_hdr->daddr;

  uint32_t src_ip_n = ntohl(source_ip.sin_addr.s_addr);
  uint32_t dest_ip_n = ntohl(dest_ip.sin_addr.s_addr);

  uint32_t old_seq = ntohl(tcp_hdr->seq);

  uint16_t dest_port = ntohs(tcp_hdr->dest);

  cout << "spa data not found -- calculating..." << endl;

  memset(bridgeSecret.key, 0, sizeof(bridgeSecret.key));
  key = getBridgeEntry(dest_ip_n, dest_port);
	
  if (key.size() <= 0) {
    cout << "KEY NOT FOUND" << endl;
  } 

  cout << "Using key: " << key << endl; 
	
  copyKey(key, bridgeSecret.key);
  spa_conn_tag = getConnectionTag(bridgeSecret, source_ip, dest_ip, tcp_hdr, ip_hdr, (time(NULL)/60));
  spa_offset = (spa_conn_tag >> 8) - (old_seq & 0x00FFFFFF); 
  setSpaData(session, spa_offset, spa_conn_tag);
}

// libnetfilter_queue code based on tutorial from Joe Kopena
static int callback(nfq_q_handle *myQueue, struct nfgenmsg *msg,
                    nfq_data *pkt, void *cbData) {
  uint32_t packet_id = 0;
  nfqnl_msg_packet_hdr *header;
  struct iphdr *ip_hdr;
  struct tcphdr *tcp_hdr;
 
  struct sockaddr_in source_ip, dest_ip;
  uint32_t old_seq, new_seq;  // Used to calculate checksum difference, and seq offset
  uint32_t spa_conn_tag = 0;

  uint32_t dest_ip_n, src_ip_n;       // Used to access
  uint16_t dest_port, src_port;       // spa_data map

  uint16_t old_chksum = 0;
  uint16_t new_chksum = 0;

  int32_t spa_offset = 0;

  pair<int32_t, uint32_t> saved_spa_data;

  TcpSession session;

  header = nfq_get_msg_packet_hdr(pkt);
  packet_id = ntohl(header->packet_id);

  // Print the payload; in copy meta mode, only headers will be included;
  // in copy packet mode, whole packet will be returned.
  char *pktData;
  int len = nfq_get_payload(pkt, &pktData);
  
  if (!len) {
    return nfq_set_verdict(myQueue, packet_id, NF_ACCEPT, 0, NULL);
  }

  memset(&source_ip, 0, sizeof(source_ip));
  memset(&dest_ip, 0, sizeof(dest_ip));

  ip_hdr = (struct iphdr *) pktData;
  tcp_hdr = (struct tcphdr *) (pktData + sizeof(struct iphdr)); 

  source_ip.sin_addr.s_addr = ip_hdr->saddr;
  dest_ip.sin_addr.s_addr = ip_hdr->daddr;

  src_ip_n = ntohl(source_ip.sin_addr.s_addr);
  dest_ip_n = ntohl(dest_ip.sin_addr.s_addr);
  src_port = ntohs(tcp_hdr->source);
  dest_port = ntohs(tcp_hdr->dest);

  cout << "Source Address: " << inet_ntoa(source_ip.sin_addr) << endl;
  cout << "Destination Address: " << inet_ntoa(dest_ip.sin_addr) << endl;
  cout << "Source Port: " << (uint32_t)src_port << endl;
  cout << "Destination Port: " << (uint32_t)dest_port << endl;
  cout << "Sequence Number: " << hex << (unsigned int)ntohl(tcp_hdr->seq) << endl;
  cout << "Acknowledgement Number: " << (unsigned int)ntohl(tcp_hdr->ack_seq) << endl;
  cout << "TCP Checksum: " << (unsigned short)ntohs(tcp_hdr->check) << dec << endl;
  
  old_seq = ntohl(tcp_hdr->seq);

  session.setData(src_ip_n, dest_ip_n, src_port, dest_port);
  saved_spa_data = findSpaData(session);
  spa_offset = saved_spa_data.first;
  spa_conn_tag = saved_spa_data.second;

  if (!spa_offset && !spa_conn_tag) {
    getTagData(session, spa_offset, spa_conn_tag, ip_hdr, tcp_hdr);
  }

  cout << "ConntectionTag: " << hex << (uint32_t)spa_conn_tag << dec << endl;
  cout << "SPA offset " << (int32_t)spa_offset << endl;

  tcp_hdr->seq = htonl(old_seq + spa_offset);

  //Only adjust ack if there is an ackseq
  if (tcp_hdr->ack) {
    tcp_hdr->ack_seq = htonl(ntohl(tcp_hdr->ack_seq) - spa_offset);
  } 

  if (tcp_hdr->syn) {
    unsigned char timestamp_byte = ((unsigned char *)(&spa_conn_tag))[0];

    //Do we need to adjust the echo value or the timestamp value?    
    if (tcp_hdr->ack) {
      setTimestamp(tcp_hdr, timestamp_byte, true, session);
    } else {
      setTimestamp(tcp_hdr, timestamp_byte, false, session);
    }
  }

  tcp_hdr->check = 0;  
  tcp_hdr->check = tcpCSum(ip_hdr, tcp_hdr);

  cout << "***" << endl;

  cout << "Sequence Number: " << hex << (unsigned int)ntohl(tcp_hdr->seq) << endl;
  cout << "Acknowledgement Number: " << (unsigned int)ntohl(tcp_hdr->ack_seq) << endl;
  cout << "TCP Checksum " << hex << ntohs(tcp_hdr->check) << dec;  
  cout << endl;
  cout << endl;

  return nfq_set_verdict(myQueue, packet_id, NF_ACCEPT, len, (unsigned char*)pktData);
}

// libnetfilter_queue code based on tutorial from Joe Kopena
void* processQueue(void *data) {
  struct nfq_handle *nfqHandle;
  struct nfq_q_handle *myQueue;
  struct nfnl_handle *netlinkHandle;

  int fd, res;
  char buf[4096];

  // Get a queue connection handle from the module
  if (!(nfqHandle = nfq_open())) {
    cerr << "Error in nfq_open()" << endl;
   return NULL;
  }

  // Unbind the handler from processing any IP packets
  // Not totally sure why this is done, or if it's necessary...
  if (nfq_unbind_pf(nfqHandle, AF_INET) < 0) {
    cerr << "Error in nfq_unbind_pf()" << endl;
  }

  // Bind this handler to process IP packets...
  if (nfq_bind_pf(nfqHandle, AF_INET) < 0) {
    cerr << "Error in nfq_bind_pf()" << endl;
  }

  // Install a callback on queue 0
  if (!(myQueue = nfq_create_queue(nfqHandle,  1, &callback, NULL))) {
    cerr << "Error in nfq_create_queue()" << endl;
    return NULL;
  }

  // Turn on packet copy mode
  if (nfq_set_mode(myQueue, NFQNL_COPY_PACKET, 0xffff) < 0) {
    cerr << "Could not set packet copy mode" << endl;
    return NULL;
  }

  netlinkHandle = nfq_nfnlh(nfqHandle);
  fd = nfnl_fd(netlinkHandle);

  while ((res = recv(fd, buf, sizeof(buf), 0)) && res >= 0) {
    nfq_handle_packet(nfqHandle, buf, res);
  }

  close(fd);
  nfq_destroy_queue(myQueue);

  nfq_close(nfqHandle);
  return 0;
}

int main(int argc, char **argv) {
  pthread_t t_process_queue, t_serve_control_port;

  signal(SIGABRT, &removeQueueRulesAndDie);
  signal(SIGTERM, &removeQueueRulesAndDie);
  signal(SIGINT, &removeQueueRulesAndDie);
  
  pthread_create( &t_process_queue, NULL, processQueue, (void*) 0);
  pthread_create( &t_serve_control_port, NULL, serveControlPort, (void*) 0);
  
  pthread_join(t_process_queue, NULL);
  pthread_join(t_serve_control_port, NULL);
}
